module Picard_11_Basics where
import Picard_11_Data_Types
import FastSort



-------------------------------------------------------------------------------
-- rec2Steps
-- For the sake of the simplicity, at the beginning both, function
--   and derivative, are given as a sequence of rectangles (which
--   represent step functions). Function and derivative take constant
--   intervals over each of the crescents. With this function for
--   each of the crescents we find such interval value.
-------------------------------------------------------------------------------
recs2Steps              :: QRecs -> QSteps
recs2Steps recs         = [rec2Step (QINotDef xf) recs | xf <- crescents]
  where
  crescents = zip nodes (tail nodes)
  nodes     = partitionFromQRecs recs
--
rec2Step                :: QStep -> QRecs -> QStep
rec2Step f []           = f
rec2Step (QIFail xf)_   = QIFail xf
rec2Step (QINotDef xf) ((x,(y1,y2)):rs)
  | isAbove xf x        = rec2Step (QIDef (xf,(y1,y1),(y2,y2))) rs
  | otherwise           = rec2Step (QINotDef xf) rs
rec2Step (QIDef (xf,yfL,yfU)) ((x,y):rs)
  | not (isAbove xf x)  = rec2Step (QIDef (xf,yfL,yfU)) rs
  | new_yfL <= new_yfU  = rec2Step (QIDef (xf,(new_yfL,new_yfL),(new_yfU,new_yfU))) rs
  | otherwise           = QIFail xf
  where
  (new_yfL,new_yfU) = head (join (fst yfL, fst yfU) y)
--
--
--------------------------------------------------------------------------------
-- partitionFromQRecs
-- Creates a partition set (from QRecs)
--
-- getXFromQRec
-- Gets both x values from a single Rec
--------------------------------------------------------------------------------
partitionFromQRecs                  :: QRecs -> [Rational]
partitionFromQRecs fs               = (msort . noDuplicates . concat) xList
  where
  xList = map getXFromQRec fs
--
getXFromQRec                        :: QRec -> [Rational]
getXFromQRec (((x1,x2),y))          = [x1,x2]
--
--
--------------------------------------------------------------------------------
-- syncFunctions
-- This takes a list of functions of type QSteps and syncronises them all
--   over the same partition set
-- Prior to this, all functions are smoothened.
--------------------------------------------------------------------------------
syncFunctions           :: [QSteps] -> [QSteps]
syncFunctions fs        = map (aux_addNodes nodes) (map smoothenFunction fs)
  where
  aux_addNodes ns f     = foldl (flip addNode2QSteps) f ns
  nodes = partitionFromQSteps (concat fs)
--
--
--------------------------------------------------------------------------------
-- partitionFromQSteps
-- Creates a partition set (from QSteps)
--
-- getXFromQStep
-- Gets both x values from a single QStep piece
--------------------------------------------------------------------------------
partitionFromQSteps                 :: QSteps -> [Rational]
partitionFromQSteps f               = (msort . noDuplicates . concat) xList
  where
  xList = map getXFromQStep f
--
--
getXFromQStep                       :: QStep -> [Rational]
getXFromQStep (QIFail (x1,x2))      = [x1,x2]
getXFromQStep (QINotDef (x1,x2))    = [x1,x2]
getXFromQStep (QIDef ((x1,x2),_,_)) = [x1,x2]
--
--
getXFromQCubes                      :: QCubes -> [Rational]
getXFromQCubes []                   = []
getXFromQCubes (((x1,x2),_,_):qs)   = x1 : x2 : (getXFromQCubes qs)
--
--
--------------------------------------------------------------------------------
-- addNode2QSteps
-- Function adds a new partition node to a function of type QSteps
--
-- pushNodeInQStep
-- Is auxiliarly function which pushes a new node in a single piece
--------------------------------------------------------------------------------
addNode2QSteps              :: Rational -> QSteps -> QSteps
addNode2QSteps _ []         = []
addNode2QSteps x [f]
  | (x<x1)                  = [QINotDef (x,x1),f]
  | (x>x2)                  = [f,QINotDef (x2,x)]
  | otherwise               = pushNodeInQStep x f
  where
  [x1,x2] = getXFromQStep f
addNode2QSteps x (f:fs)
  | (x<x1)                  = (QINotDef (x,x1)) : f : fs
  | (x<=x2)                 = (pushNodeInQStep x f) ++ fs
  | otherwise               = f : (addNode2QSteps x fs)
  where
  [x1,x2] = getXFromQStep f
--
--
pushNodeInQStep :: Rational -> QStep -> [QStep]
pushNodeInQStep x (QIFail (x1,x2))
  | (x==x1) || (x==x2)      = [QIFail (x1,x2)]
  | (x>x1)  && (x<x2)       = [QIFail (x1,x),QIFail (x,x2)]
  | otherwise               = error "Error 271 (pushNodeInQStep)"
pushNodeInQStep x (QINotDef (x1,x2))
  | (x==x1) || (x==x2)      = [QINotDef (x1,x2)]
  | (x>x1)  && (x<x2)       = [QINotDef (x1,x),QINotDef (x,x2)]
  | otherwise               = error "Error 272 (pushNodeInQStep)"
pushNodeInQStep x (QIDef ((x1,x2),(yL1,yL2),(yU1,yU2)))
  | (x==x1) || (x==x2)      = [QIDef ((x1,x2),(yL1,yL2),(yU1,yU2))]
  | (x1<x) && (x<x2)        = [QIDef ((x1,x),(yL1,yLnew),(yU1,yUnew)),
                               QIDef ((x,x2),(yLnew,yL2),(yUnew,yU2))]
  | otherwise               = error "Error 273 (pushNodeInQStep)"
  where
  yLnew = (yL2-yL1)*x/(x2-x1) + (x2*yL1-x1*yL2)/(x2-x1)
  yUnew = (yU2-yU1)*x/(x2-x1) + (x2*yU1-x1*yU2)/(x2-x1)
--
--
--------------------------------------------------------------------------------
-- smoothenFunction
--
-- It's quite possible that a function has too many pieces. Here, we try to
--   make the number of pieces as small as possible.
--------------------------------------------------------------------------------
smoothenFunction     :: QSteps -> QSteps
smoothenFunction []  = []
smoothenFunction [f] = [f]
smoothenFunction ((QIFail xf):(QIFail xh):fs)
  | cond             = smoothenFunction (fnew:fs)
  | otherwise        = error "Error 511 (smoothenFunction)"
  where
  cond = (snd xf == fst xh)
  fnew = QIFail (fst xf,snd xh)
smoothenFunction ((QIFail xf):fs)
                     = (QIFail xf) : (smoothenFunction fs)
smoothenFunction ((QINotDef xf):(QINotDef xh):fs)
  | cond             = smoothenFunction (fnew:fs)
  | otherwise        = error "Error 512 (smoothenFunction)"
  where
  cond = (snd xf == fst xh)
  fnew = QINotDef (fst xf,snd xh)
smoothenFunction ((QINotDef xf):fs)
                     = (QINotDef xf) : (smoothenFunction fs)
smoothenFunction ((QIDef (xf,yfL,yfU)):(QIDef (xh,yhL,yhU)):fs)
  | cond             = smoothenFunction (fnew:fs)
  | otherwise        = (QIDef (xf,yfL,yfU)) : (smoothenFunction ((QIDef (xh,yhL,yhU)):fs))
  where
  (xf1,xf2) = xf; (xh1,xh2) = xh;
  (yfL1,yfL2) = yfL; (yfU1,yfU2) = yfU;
  (yhL1,yhL2) = yhL; (yhU1,yhU2) = yhU;
  cond  = cond1 && cond2 && cond3
  cond1 = snd xf == fst xh
  cond2 = yfL2 == (yfL1 + (xf2-xf1)*(yhL2-yfL1)/(xh2-xf1))
  cond3 = yfU2 == (yfU1 + (xf2-xf1)*(yhU2-yfU1)/(xh2-xf1))
  fnew  = QIDef ((xf1,xh2),(yfL1,yhL2),(yfU1,yhU2))
smoothenFunction ((QIDef (xf,yfL,yfU)):fs)
                     = (QIDef (xf,yfL,yfU)) : (smoothenFunction fs)
--
--
-------------------------------------------------------------------------------
-- linesIntersection
--   Computes coordinates of the intersection of two lines, both given as
--   a pair of points. It produces an error message if their intersection
--   is not a singleton set.
-------------------------------------------------------------------------------
linesIntersection ((x11,y11),(x12,y12)) ((x21,y21),(x22,y22))
  | a1 /= a2  = (x,y)
  | b1 /= b2  = error "Error 301 (lines are parallel)"
  | otherwise = error "Error 302 (two lines are same)"
  where
  x  = (b2-b1) / (a1-a2)
  y  = (a1*b2-a2*b1) / (a1-a2)
  a1 = (y12-y11) / (x12-x11)
  a2 = (y22-y21) / (x22-x21)
  b1 = y11-a1*x11
  b2 = y21-a2*x21
--
--
-------------------------------------------------------------------------------
-- noDuplicates - from a list of elements produces another list with only
--                one occurrence of each element from the input list.
-------------------------------------------------------------------------------
noDuplicates        :: Eq a => [a] -> [a]
noDuplicates []     = []
noDuplicates (x:xs)
       | elem x xs  = noDuplicates xs
       | otherwise  = x : noDuplicates xs
--
--
-------------------------------------------------------------------------------
-- 'Domain-theoretic functions'
-- isAbove  = checks if (a,b) is subset of (c,d)
-- join     = finds the greatest interval which is subset of both, (a,b)
--            and (c,d). If the first coordinate of the result is greater
--            of the second coordinate, we have bottom (i.e. error message)
-------------------------------------------------------------------------------
isAbove             :: Ord t => (t,t) -> (t,t) -> Bool
isAbove (a,b) (c,d) = (a>=c) && (b<=d)
--
--
join                :: (Show t, Ord t) => (t,t) -> (t,t) -> [(t,t)]
join (a,b) (c,d)
  | x <= y          = [(x, y)]
  | otherwise       = [] -- error "Error (lub of two intervals is bottom)"
  where
  (x,y) = (max a c, min b d)
--
--
-------------------------------------------------------------------------------
-- flipLU2UL
-------------------------------------------------------------------------------
flipLU2UL                        :: QSteps -> QSteps
flipLU2UL []                     = []
flipLU2UL ((QINotDef x):fs)      = (QINotDef x)      : (flipLU2UL fs)
flipLU2UL ((QIFail x):fs)        = (QIFail x)        : (flipLU2UL fs)
flipLU2UL ((QIDef (x,yL,yU)):fs) = (QIDef (x,yU,yL)) : (flipLU2UL fs)
