module Picard_11_Update where
import Picard_11_Data_Types
import Picard_11_Basics
import FastSort


--
--
--------------------------------------------------------------------------------
-- updateDerivative
-- This updates the information about the derivative from function (fs) and
--   vector field (vfs). The vector field is give as a finate lub of simple
--   step functions (with two arguments).
--
-- whereAbove
-- This checks where, over which subinterval, a single QStep piece
--   is within bounds (lower,upper)=(\underline(b_i),\overline(b_i))
--
-- d_(ij) \subset d_j, with
-- d_(ij)^o = ((g_j^-)^(-1)(\underline(b_i),\infty) \cap
--            ((g_j^+)^(-1)(-\infty,\overline(b_i))
--------------------------------------------------------------------------------
--
--
updateDerivative                :: QCubes -> QSteps -> QRecs
updateDerivative vfs fs         = concat [ updateDerivative_1 vf f | vf<-vfs, f<-fsNew ]
  where
  xList = (msort . noDuplicates) (getXFromQCubes vfs)
  fsNew = last (scanl (flip addNode2QSteps) fs xList)
--
--
updateDerivative_1              :: QCube -> QStep -> [QRec]
updateDerivative_1 _ (QIFail x) = []
updateDerivative_1 (a,b,c) (QINotDef x)
   | True                       = []
updateDerivative_1 (a,b,c) (QIDef (x,yL,yU))
  | length (join a x) == 0      = []
  | length xij == 0             = []
  | length ax == 0              = []
  | otherwise                   = [(head ax,c)]
  where
  xij = whereAbove (QIDef (x,yL,yU)) b
  ax  = join a (head xij)
--
--
whereAbove                  :: QStep -> (Rational,Rational) -> [QInterval]
whereAbove (QIFail x) _     = []
whereAbove (QINotDef x) _   = []
whereAbove (QIDef (x,yL,yU)) (lbound,ubound)
  | length region1 == 0     = []
  | length region2 == 0     = []
  | left > right            = []
  | otherwise               = [(left,right)]
  where
  (x1,x2) = x; (yL1,yL2) = yL; (yU1,yU2) = yU
  xpL     = fst (linesIntersection ((x1,yL1),(x2,yL2)) ((x1,lbound),(x2,lbound)))
  xpU     = fst (linesIntersection ((x1,yU1),(x2,yU2)) ((x1,ubound),(x2,ubound)))
  left    = max (fst (head region1)) (fst (head region2))
  right   = min (snd (head region1)) (snd (head region2))
  region1 = case (yL1>=lbound,yL2>=lbound) of
            (True,True)   -> [(x1,x2)]
            (True,False)  -> [(x1,xpL)]
            (False,True)  -> [(xpL,x2)]
            (False,False) -> []
  region2 = case (yU1<=ubound,yU2<=ubound) of
            (True,True)   -> [(x1,x2)]
            (True,False)  -> [(x1,xpU)]
            (False,True)  -> [(xpU,x2)]
            (False,False) -> []
--
--
--------------------------------------------------------------------------------
-- updateFunction
-- Updates the information about the function (fs) with the information about
--   the derivative (gs)
--
-- updateSCFunction
-- This flips ((lower,upper) -> (upper,lower)) the information about the
--   derivative and then updates the function. We need this to check
--   strong consistency
--------------------------------------------------------------------------------
updateFunction                        :: QSteps -> QSteps -> QSteps
updateFunction fs gs                  = uf_R2L fs4 gs4
  where
  [fs2,gs2] = syncFunctions [fs,gs]
  fs3 = uf_L2R fs2 gs2
  [fs4,gs4] = syncFunctions [fs3,gs2]
--
--
updateSCFunction                      :: QSteps -> QSteps -> QSteps
updateSCFunction fs gs                = uf_R2L fs4 gs4
  where
  [fs2,gs2] = syncFunctions [fs,flipLU2UL gs]
  fs3 = uf_L2R fs2 gs2
  [fs4,gs4] = syncFunctions [fs3,gs2]
--
--
--------------------------------------------------------------------------------
-- areQStepsStronglyConsistent
-- It does exactly what it says on the tin
--------------------------------------------------------------------------------
areQStepsStronglyConsistent :: QSteps -> QSteps -> Bool
areQStepsStronglyConsistent fs gs       = foldl (&&) True clist
  where
  fSC       = updateSCFunction fs gs
  [f2,fSC2] = syncFunctions [fs,fSC]
  clist     = zipWith isQStepContained fSC2 f2
--
--
--------------------------------------------------------------------------------
-- areQStepsConsistent
-- Checks if fs (as function) and gs (as derivative) are consistent
--------------------------------------------------------------------------------
areQStepsConsistent                     :: QSteps -> QSteps -> Bool
areQStepsConsistent fs gs               = foldl (&&) True wdlist
  where
  fupdate = updateFunction fs gs
  wdlist  = map isQStepWellDefined fupdate
--
--
--------------------------------------------------------------------------------
-- isQStepContained
-- For two QStep pieces over the same interval, it checks if the first one is
--   inside/refining the second one
--------------------------------------------------------------------------------
isQStepContained                        :: QStep -> QStep -> Bool
isQStepContained (QIFail xf) _          = False
isQStepContained _ (QIFail xh)          = False
isQStepContained (QINotDef xf) _        = error "Error 171 (isQStepContained)" -- True
isQStepContained _ (QINotDef xh)        = True
isQStepContained (QIDef (xf,yfL,yfU)) (QIDef (xh,yhL,yhU))
  | xf /= xh                            = error "Error 172 (isQStepContained)"
  | otherwise                           = cond1 && cond2
  where
  cond1 = (yfL>=yhL) && (yfL<=yhU)
  cond2 = (yfU>=yhL) && (yfU<=yhU)
--
--
--------------------------------------------------------------------------------
-- isQStepWellDefined
-- This checks if a single piece (QStep) is properly defined, that is, if
--   lower function is really (visually) below the upper
--------------------------------------------------------------------------------
isQStepWellDefined                      :: QStep -> Bool
isQStepWellDefined (QIFail xf)          = False
isQStepWellDefined (QINotDef xf)        = True
isQStepWellDefined (QIDef (xf,yfL,yfU)) = yfL <= yfU
--
--
-------------------------------------------------------------------------------
-- uf_R2L = update function (right-2-left-step)
--
-- Input functions, fs & gs, must be over the same partition,
--   have the same number of elements, etc.
-------------------------------------------------------------------------------
uf_R2L                                :: QSteps -> QSteps -> QSteps
uf_R2L fs gs
  | length fs /= length gs            = error "Error 440 (uf_R2L)"
  | otherwise                         = concat (map fst upList)
  where
  upList = scanr (flip singleStep) ([],CarryN) (zip fs gs)
  singleStep (aux,c) (f,g) = uf_R2L_1 c (f,g)
--
--
uf_R2L_1 :: Carried -> (QStep,QStep) -> ([QStep],Carried)
uf_R2L_1 CarryN ((QIFail xf),_)       = ([QIFail xf],CarryN)
uf_R2L_1 CarryN ((QINotDef xf),_)     = ([QINotDef xf],CarryN)
uf_R2L_1 CarryN ( (QIDef (xf,yfL,yfU)),(QIDef (xg,ygL,ygU)) )
  | xf == xg                          = ([QIDef (xf,(new_yfL1,yfL2),(new_yfU1,yfU2))],
                                        CarryY new_yfL1 new_yfU1)
  | otherwise                         = error "Error 441 (uf_R2L_1)"
  where
  (x1,x2)  = xf; (yfL1,yfL2) = yfL; (yfU1,yfU2) = yfU;
  new_yfL1 = maximum [yfL2-(fst ygU)*(x2-x1), yfL1]
  new_yfU1 = minimum [yfU2-(fst ygL)*(x2-x1), yfU1]
uf_R2L_1 CarryN _                     = error "Error 442 (uf_R2L_1)"
--
uf_R2L_1 (CarryY _ _) ((QIFail xf),_) = ([QIFail xf],CarryN)
uf_R2L_1 (CarryY pyL pyU) ( (QINotDef xf),(QIDef (xg,ygL,ygU)) )
  | xf == xg                          = ([QIDef (xf,(new_yfL1,pyL),(new_yfU1,pyU))],
                                        CarryY new_yfL1 new_yfU1)
  | otherwise                         = error "Error 443 (uf_R2L_1)"
  where
  new_yfL1 = pyL-(fst ygU)*(snd xf - fst xf)
  new_yfU1 = pyU-(fst ygL)*(snd xf - fst xf)
uf_R2L_1 (CarryY pyL pyU) ( (QIDef (xf,yfL,yfU)),(QIDef (xg,ygL,ygU)) )
  | xf == xg                          = (res2, CarryY ycL ycU)
  | otherwise                         = error "Error 444 (uf_R2L_1)"
  where
  QIDef (_,(ycL,_),(ycU,_)) = head res2
  res2     = reverse (pushUpperSegment uSegment res1)
  res1     = pushLowerSegment lSegment [QIDef (xf,yfL,yfU)]
  (x1,x2)  = xf; (yfL1,yfL2) = yfL; (yfU1,yfU2) = yfU;
  new_yfL2 = maximum [pyL,yfL2]
  new_yfU2 = minimum [pyU,yfU2]
  new_yfL1 = new_yfL2-(fst ygU)*(x2-x1)
  new_yfU1 = new_yfU2-(fst ygL)*(x2-x1)
  lSegment = (xf,(new_yfL1,new_yfL2))
  uSegment = (xf,(new_yfU1,new_yfU2))
uf_R2L_1 (CarryY _ _) _               = error "Error 445 (uf_R2L_1)"
--
--
-------------------------------------------------------------------------------
-- uf_L2R = update function (left-2-right-step)
--
-- Input functions, fs & gs, must be over the same partition,
--   have the same number of elements, etc.
-------------------------------------------------------------------------------
uf_L2R                                :: QSteps -> QSteps -> QSteps
uf_L2R fs gs
  | length fs /= length gs            = error "Error 420 (uf_L2R)"
  | otherwise                         = concat (map fst upList)
  where
  upList = scanl singleStep ([],CarryN) (zip fs gs)
  singleStep (aux,c) (f,g) = uf_L2R_1 c (f,g)
--
--
uf_L2R_1 :: Carried -> (QStep,QStep)  -> ([QStep],Carried)
uf_L2R_1 CarryN ((QIFail xf),_)       = ([QIFail xf],CarryN)
uf_L2R_1 CarryN ((QINotDef xf),_)     = ([QINotDef xf],CarryN)
uf_L2R_1 CarryN ( (QIDef (xf,yfL,yfU)),(QIDef (xg,ygL,ygU)) )
  | xf == xg                          = ([QIDef (xf,(yfL1,new_yfL2),(yfU1,new_yfU2))],
                                        CarryY new_yfL2 new_yfU2)
  | otherwise                         = error "Error 421 (uf_L2R_1)"
  where
  (x1,x2)  = xf; (yfL1,yfL2) = yfL; (yfU1,yfU2) = yfU;
  new_yfL2 = maximum [yfL1+(fst ygL)*(x2-x1), yfL2]
  new_yfU2 = minimum [yfU1+(fst ygU)*(x2-x1), yfU2]
uf_L2R_1 CarryN _                     = error "Error 422 (uf_L2R_1)"
--
uf_L2R_1 (CarryY _ _) ((QIFail xf),_) = ([QIFail xf],CarryN)
uf_L2R_1 (CarryY pyL pyU) ( (QINotDef xf),(QIDef (xg,ygL,ygU)) )
  | xf == xg                          = ([QIDef (xf,(pyL,new_yfL2),(pyU,new_yfU2))],
                                        CarryY new_yfL2 new_yfU2)
  | otherwise                         = error "Error 423 (uf_L2R_1)"
  where
  new_yfL2 = pyL+(fst ygL)*(snd xf - fst xf)
  new_yfU2 = pyU+(fst ygU)*(snd xf - fst xf)
uf_L2R_1 (CarryY pyL pyU) ( (QIDef (xf,yfL,yfU)),(QIDef (xg,ygL,ygU)) )
  | xf == xg                          = (res2, CarryY ycL ycU)
  | otherwise                         = error "Error 424 (uf_L2R_1)"
  where
  QIDef (_,(_,ycL),(_,ycU)) = last res2
  res2     = pushUpperSegment uSegment res1
  res1     = pushLowerSegment lSegment [QIDef (xf,yfL,yfU)]
  (x1,x2)  = xf; (yfL1,yfL2) = yfL; (yfU1,yfU2) = yfU;
  new_yfL1 = maximum [pyL,yfL1]
  new_yfU1 = minimum [pyU,yfU1]
  new_yfL2 = new_yfL1+(fst ygL)*(x2-x1)
  new_yfU2 = new_yfU1+(fst ygU)*(x2-x1)
  lSegment = (xf,(new_yfL1,new_yfL2))
  uSegment = (xf,(new_yfU1,new_yfU2))
uf_L2R_1 (CarryY _ _) _               = error "Error 425 (uf_L2R_1)"
--
--
-------------------------------------------------------------------------------
-- pushLowerSegment
-- pushUpperSegment
-------------------------------------------------------------------------------
pushLowerSegment :: (QInterval,QInterval) -> [QStep] -> [QStep]
pushLowerSegment (x,y) fs = case (length fs) of
  1 -> pushLowerSegment_1 (x,y) fs
  2 -> pushLowerSegment_2 (x,y) fs
  _ -> error "Error 250 (pushLowerSegment)"
--
pushLowerSegment_1 :: (QInterval,QInterval) -> [QStep] -> [QStep]
pushLowerSegment_1 (x,y) [QIDef (xf,yL,yU)]
  | x /= xf                = error "Error 251 (pushLowerSegment_1)"
  | (y1>=yL1) && (y2>=yL2) = [QIDef (x,y,yU)]
  | (y1<=yL1) && (y2<=yL2) = [QIDef (x,yL,yU)]
  | (y1>=yL1) && (y2<=yL2) = [QIDef ((x1,xp),( y1,yLp),(yU1,yUp)),
                              QIDef ((xp,x2),(yLp,yL2),(yUp,yU2))]
  | (y1<=yL1) && (y2>=yL2) = [QIDef ((x1,xp),(yL1,yLp),(yU1,yUp)),
                              QIDef ((xp,x2),(yLp, y2),(yUp,yU2))]
  | otherwise              = error "Error 252 (pushLowerSegment_1)"
  where
  (x1,x2)=x; (y1,y2)=y; (yL1,yL2)=yL; (yU1,yU2)=yU
  (xp,yLp) = linesIntersection ((x1,y1),(x2,y2)) ((x1,yL1),(x2,yL2))
  yUp = yU1 + (xp-x1)*(yU2-yU1)/(x2-x1)
pushLowerSegment_1 _ _     = error "Error 253 (pushLowerSegment_1)"
--
pushLowerSegment_2 :: (QInterval,QInterval) -> [QStep] -> [QStep]
pushLowerSegment_2 (x,y) [f] = error "Error 254 (pushLowerSegment_2)"
pushLowerSegment_2 (x,y) fs
  | cond                     = res1 ++ res2 
  | otherwise                = error "Error 255 (pushLowerSegment_2)"
  where
  cond = (xf1==x1) && (xf2==xf3) && (xf4==x2)
  (x1,x2)=x; (y1,y2)=y
  [xf1,xf2] = getXFromQStep (fs!!0)
  [xf3,xf4] = getXFromQStep (fs!!1)
  yp = y1 + (xf2-x1)*(y2-y1)/(x2-x1)
  res1 = pushLowerSegment_1 ((x1,xf2),(y1,yp)) (take 1 fs)
  res2 = pushLowerSegment_1 ((xf3,x2),(yp,y2)) (drop 1 fs)
--
--
pushUpperSegment :: (QInterval,QInterval) -> [QStep] -> [QStep]
pushUpperSegment (x,y) fs = case (length fs) of
  1 -> pushUpperSegment_1 (x,y) fs
  2 -> pushUpperSegment_2 (x,y) fs
  _ -> error "Error 260 (pushUpperSegment)"
--
pushUpperSegment_1 :: (QInterval,QInterval) -> [QStep] -> [QStep]
pushUpperSegment_1 (x,y) [QIDef (xf,yL,yU)]
  | x /= xf                = error "Error 261 (pushUpperSegment_1)"
  | (y1<=yU1) && (y2<=yU2) = [QIDef (x,yL,y)]
  | (y1>=yU1) && (y2>=yU2) = [QIDef (x,yL,yU)]
  | (y1<=yU1) && (y2>=yU2) = [QIDef ((x1,xp),(yL1,yLp),( y1,yUp)),
                              QIDef ((xp,x2),(yLp,yL2),(yUp,yU2))]
  | (y1>=yU1) && (y2<=yU2) = [QIDef ((x1,xp),(yL1,yLp),(yU1,yUp)),
                              QIDef ((xp,x2),(yLp,yL2),(yUp, y2))]
  | otherwise              = error "Error 262 (pushUpperSegment_1)"
  where
  (x1,x2)=x; (xf1,xf2)=xf
  (y1,y2)=y; (yL1,yL2)=yL; (yU1,yU2)=yU
  (xp,yUp) = linesIntersection ((x1,y1),(x2,y2)) ((x1,yU1),(x2,yU2))
  yLp = yL1 + (xp-x1)*(yL2-yL1)/(x2-x1)
pushUpperSegment_1 _ _     = error "Error 263 (pushUpperSegment_1)"
--
pushUpperSegment_2 :: (QInterval,QInterval) -> [QStep] -> [QStep]
pushUpperSegment_2 (x,y) [f] = error "Error 264 (pushUpperSegment_2)"
pushUpperSegment_2 (x,y) fs
  | cond                     = res1 ++ res2 
  | otherwise                = error "Error 265 (pushUpperSegment_2)"
  where
  cond = (xf1==x1) && (xf2==xf3) && (xf4==x2)
  (x1,x2)=x; (y1,y2)=y
  [xf1,xf2] = getXFromQStep (fs!!0)
  [xf3,xf4] = getXFromQStep (fs!!1)
  yp = y1 + (xf2-x1)*(y2-y1)/(x2-x1)
  res1 = pushUpperSegment_1 ((x1,xf2),(y1,yp)) (take 1 fs)
  res2 = pushUpperSegment_1 ((xf3,x2),(yp,y2)) (drop 1 fs)

